import torch
import torch.nn as nn
import torch.nn.functional as F

import numpy as np
import matplotlib.pyplot as plt
# import os

torch.cuda.empty_cache()
π = np.pi
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

#############################################################
### BASIC STUFF #############################################
#############################################################

def nn_size(nn: nn.Module): return sum([p.numel() for p in nn.parameters()]) # get param count

def nn_freeze(model: nn.Module):
	model.eval() # TODO : prevent going back to train mode
	for param in model.parameters(): param.requires_grad = False
	for l in model.children(): nn_freeze(l)
	return model

def nn_unfreeze(model: nn.Module):
	for param in model.parameters(): param.requires_grad = True
	for l in model.children(): nn_freeze(l)
	return model

def optim_lr(optimizer: torch.optim.Optimizer, lr: float):
	for g in optimizer.param_groups: g['lr'] = lr

def nn_dropper(n: int): return modulize(lambda x : x[:,:n])

def nn_permute_and_drop(input_size: int, output_size: int):
	perm = torch.randperm(input_size)
	return modulize(lambda x : x[:,perm][:,:output_size])

#############################################################
### USEFUL FUNCTIONS ########################################
#############################################################

### interpolations ########################
def nearest(t,v1,v2): λ = (t>0.5).float(); return v1*(1-λ) + v2*λ
def lerp(t,v1,v2):                         return v1*(1-t) + v2*t
def smoothstep(t,v1,v2): λ = t*t*(3-2*t);  return v1*(1-λ) + v2*λ

### sampler ###############################
def sampler1d_repeat(tex,x,interpolation=lerp):
	l = len(tex)
	X=x*l # so that x ∈ [0,1] wil cover the whole texture
	Xi = X.int() - (X<=0)*1
	t = X-Xi
	return interpolation(t, tex[Xi % l], tex[(Xi + 1) % l])

def sampler1d_clamp(tex,x,interpolation=lerp):
	l = len(tex)
	X=x*l # so that x ∈ [0,1] wil cover the whole texture
	Xi = X.int() - (X<=0)*1
	t = X-Xi
	return interpolation(t, tex[torch.clamp(Xi,0,l-1)], tex[torch.clamp(Xi+1,0,l-1)])

def sampler2d_nearest_clamp(tex, coord):
	c, h, w = tex.shape
	# remap coord from [0,1]x[0,1] to [0,h]x[0,w]
	coord[:, 0] *= h-1
	coord[:, 1] *= w-1
	coord = torch.round(coord).long()
	coord_x = torch.clamp(coord[:, 1],0,w-1) # repeat last pixel
	coord_y = torch.clamp(coord[:, 0],0,h-1) # repeat last pixel
	pixels = tex[:, coord_y, coord_x]
	return pixels.permute(1,0)

### polynoms ##############################
def poly(coeff): # each components get a polynome applied
	N = range(len(coeff))
	return lambda x: sum([coeff[n]*(x**n) for n in N])

### fake gradient | cantor ################
def cantor_function(x,n=6):
	if n==0: return x
	return (
		(((0 <=x) & (x<=1/3)) * cantor_function(3*x,n-1)/2) +
		((1/3< x) & (x< 2/3)) * (1/2) +
		(2/3 <=x)             * (1/2+cantor_function(3*x-2,n-1)/2)
	)

cantor_array = None
def cantor():
	global cantor_array
	if cantor_array==None : cantor_array = cantor_function(torch.linspace(0,1,3**4+1),4).to(device)
	return modulize(lambda x: sampler1d_clamp(cantor_array,x,nearest))

def cantor2():
	global cantor_array
	if cantor_array==None : cantor_array = cantor_function(torch.linspace(0,1,3**4+1),4).to(device)
	return modulize(gradiator(
		lambda x: torch.floor(x) + sampler1d_repeat(cantor_array,x,nearest),
		lambda x,grad: grad                  # fake the gradient to identity
	))

### fake gradient | sigmoid staircase #####
### fake gradient | sigmoid sinus #########
### fake gradient | binary sinus ##########

### heterogenous ##########################
def multi(σs):
	def multi_(x):
		σcount = len(σs)
		out = torch.zeros_like(x)
		for (i,σ) in enumerate(σs):
			index = torch.arange(i,len(x)-1,σcount)
			out[index] = σ(x[index])
		return out
	return modulize(multi_)

def random_function_light(dim_in=196,dim_out=196,device=device):
	L1 = nn_freeze(nn.Linear(dim_in,dim_in)).to(device)
	L2 = nn_freeze(nn.Linear(dim_in,dim_out)).to(device)
	L3 = nn_freeze(nn.Linear(dim_in,dim_out)).to(device)
	L2.weight.data = torch.relu(L2.weight.data)
	def f(x):
		x = torch.sigmoid(L2(torch.sin(L1(x)) + torch.sin(torch.relu(L2(x))))) + torch.exp(-(L3(torch.sin(x))**2))
		# x = torch.sigmoid(torch.sin(L1(x)) + torch.relu(L2(x))) + torch.exp(-(L3(torch.sin(x))**2))
		# x = L1(x) + torch.relu(L2(x))
		return torch.sin(x)
		# return torch.sigmoid(x)
	# φ = (1 + np.sqrt(5))/2
	# # def fractale(x): return f(x) + f(x*2)/2 + f(x*4)/4 + f(x*8)/8
	# def fractale(x): return f(x/φ) + f(x+φ*π) + f((x+π)*φ)/2 + f(-(x-2*φ)*φ*π)/4 + f(x*φ*φ*φ)/8
	return modulize(f).to(device)

def random_function_lighter(dim=196,device=device):
	G = torch.zeros((dim,dim)).to(device)
	for i in range(2):
		G += torch.diag(torch.randn(dim-i)/(i+1),+i).to(device)
		G += torch.diag(torch.randn(dim-i)/(i+1),-i).to(device)
	L1 = nn_freeze(nn.Linear(dim,dim)).to(device)
	# L1.weight.data = L1.weight.data * G
	def f(x):
		x = L1(x)
		# return x
		return torch.relu(x)
		# return torch.sin(x)
		# return torch.sigmoid(x)
	φ = (1 + np.sqrt(5))/2
	# def fractale(x): return f(x) + f(x*2)/2 + f(x*4)/4 + f(x*8)/8
	def fractale(x): return f(x/φ) + f(x+φ*π) + f((x+π)*φ)/2 + f(-(x-2*φ)*φ*π)/4 + f(x*φ*φ*φ)/8
	# return modulize(fractale).to(device)
	return modulize(f).to(device)

